//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "StdAfx.h"
#include "LastCommonRouteFinder.h"
#include "Pinger.h"
#include "EMule.h"


LastCommonRouteFinder::LastCommonRouteFinder() {
 	lastCommonHost = 0;
	lastCommonTTL = 0;
	hostToPing = 0;

    needMoreHosts = false;
	run = true;

    newTraceRouteHostEvent = new CEvent(0, 0);

}

LastCommonRouteFinder::~LastCommonRouteFinder() {
    delete newTraceRouteHostEvent;
}

bool LastCommonRouteFinder::AddHostsToCheck(CTypedPtrList<CPtrList, CServer*> &list) {
    if(needMoreHosts) {
        addHostLocker.Lock();
        
        if(needMoreHosts) {
            if(list.GetCount() >= 10) {
                hostsToTraceRoute.RemoveAll();

                uint32 startPos = rand()/(RAND_MAX/list.GetCount());

                POSITION pos = list.GetHeadPosition();
                for(uint32 skipCounter = startPos; skipCounter < (uint32)list.GetCount() && pos != NULL; skipCounter++) {
                    list.GetNext(pos);
                }

                uint32 tryCount = 0;
                while(pos != NULL && hostsToTraceRoute.GetCount() < 10 && tryCount <= (uint32)list.GetCount()) {
                    tryCount++;
                    CServer* server = list.GetNext(pos);

                    uint32 ip = server->GetIP();

                    if(IsGoodIP(ip, true)) {
                        hostsToTraceRoute.AddTail(ip);
                    }

                    if(pos == NULL) {
                        POSITION pos = list.GetHeadPosition();
                    }
                }
            }

            if(hostsToTraceRoute.GetCount() >= 10 || !run) {
                needMoreHosts = false;

                // Signal that there's hosts to fetch.
                newTraceRouteHostEvent->SetEvent();

                addHostLocker.Unlock();
                return true; // got enough hosts
            } else {
                addHostLocker.Unlock();
                return false; // didn't get enough hosts
            }
        } else {
            addHostLocker.Unlock();
            return true; // allready got enough hosts, don't need more
        }
    } else {
        return true; // allready got enough hosts, don't need more
    }
}

bool LastCommonRouteFinder::AddHostsToCheck(CTypedPtrList<CPtrList, CUpDownClient*> &list) {
    if(needMoreHosts) {
        addHostLocker.Lock();
        
        if(needMoreHosts) {
            if(list.GetCount() >= 10) {
                hostsToTraceRoute.RemoveAll();

                uint32 startPos = rand()/(RAND_MAX/list.GetCount());

                POSITION pos = list.GetHeadPosition();
                for(uint32 skipCounter = startPos; skipCounter < (uint32)list.GetCount() && pos != NULL; skipCounter++) {
                    list.GetNext(pos);
                }

                uint32 tryCount = 0;
                while(pos != NULL && hostsToTraceRoute.GetCount() < 10 && tryCount <= (uint32)list.GetCount()) {
                    tryCount++;
                    CUpDownClient* client = list.GetNext(pos);

                    uint32 ip = client->GetIP();

                    if(IsGoodIP(ip, true)) {
                        hostsToTraceRoute.AddTail(ip);
                    }

                    if(pos == NULL) {
                        POSITION pos = list.GetHeadPosition();
                    }
                }
            }

            if(hostsToTraceRoute.GetCount() >= 10 || !run) {
                needMoreHosts = false;

                // Signal that there's hosts to fetch.
                newTraceRouteHostEvent->SetEvent();

                addHostLocker.Unlock();
                return true; // got enough hosts
            } else {
                addHostLocker.Unlock();
                return false; // didn't get enough hosts
            }
        } else {
            addHostLocker.Unlock();
            return true; // allready got enough hosts, don't need more
        }
    } else {
        return true; // allready got enough hosts, don't need more
    }
}

/**
 * @return always returns 0.
 */
bool LastCommonRouteFinder::Find() {
	Pinger pinger;

	int tries = 0;

	// retry loop. enabled will be set to false in end of this loop, if to many failures (tries too large)
	bool foundLastCommonHost = false;

	hostsToTraceRoute.RemoveAll();

	bool atLeastOnePingSucceded = false;
	int traceRouteTries = 0;
	while(run && foundLastCommonHost == false && traceRouteTries < 3 && hostsToTraceRoute.GetCount() < 10) {
		traceRouteTries++;

		lastCommonHost = 0;

		theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Try #%i. Collecting hosts...", traceRouteTries);

		addHostLocker.Lock();
		needMoreHosts = true;
		addHostLocker.Unlock();

		// wait for hosts to traceroute
		newTraceRouteHostEvent->Lock();

		theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Got enough hosts. Listing the hosts that will be tracerouted:");

		POSITION pos = hostsToTraceRoute.GetHeadPosition();
		int counter = 0;
		while(pos != NULL) {
			counter++;
			uint32 hostToTraceRoute = hostsToTraceRoute.GetNext(pos);
			IN_ADDR stDestAddr;
			stDestAddr.s_addr = hostToTraceRoute;
			theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Host #%i: %s", counter, inet_ntoa(stDestAddr));
		}

		// find the last common host, using traceroute

		theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Starting traceroutes to find last common host.");

		bool failed = false;

		uint32 curHost = 0;
		for(uint32 ttl = 1; run && (curHost != 0 && ttl <= 64 || curHost == 0 && ttl < 5) && foundLastCommonHost == false && failed == false; ttl++) {
                        theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Pinging for TTL %i...", ttl);
			curHost = 0;
			PingStatus pingStatus;

                        uint32 lastSuccedingPingAddress = 0;
			bool firstLoop = true;
			pos = hostsToTraceRoute.GetHeadPosition();
			while(run && failed == false && pos != NULL &&
				(
				firstLoop ||
				pingStatus.success && pingStatus.destinationAddress == curHost ||
				pingStatus.success == false && pingStatus.error == IP_REQ_TIMED_OUT
				)
				) {

					firstLoop = false;

                                lastSuccedingPingAddress = 0;

					POSITION lastPos = pos;

                                // this is the current address we send ping to, in loop below.
                                // PENDING: Don't confuse this with curHost, which is unfortunately almost
                                // the same name. Will rename one of these variables as soon as possible, to
                                // get more different names.
					uint32 curAddress = hostsToTraceRoute.GetNext(pos);

					pingStatus.success = false;
					for(int counter = 0; run && counter < 3 && pingStatus.success == false; counter++) {
						pingStatus = pinger.Ping(curAddress, ttl, true);
						if(pingStatus.success == false && counter < 3-1) {
							IN_ADDR stDestAddr;
							stDestAddr.s_addr = curAddress;
							theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Failure #%i to ping host! (TTL: %i IP: %s error: %i). Sleeping 1 sec before retry. Error info follows.", counter+1, ttl, inet_ntoa(stDestAddr), pingStatus.error);
							pinger.PIcmpErr(pingStatus.error);

							Sleep(1000);

						}
					}

					if(pingStatus.success == true && pingStatus.status != IP_SUCCESS) {
						if(curHost == 0) {
							curHost = pingStatus.destinationAddress;
						}
						atLeastOnePingSucceded = true;
                                                lastSuccedingPingAddress = curAddress;
					} else {
						// failed to ping this host for some reason.
						// Or we reached the actual host we are pinging. We don't want that, since it is too close.
						// Remove it.
						IN_ADDR stDestAddr;
						stDestAddr.s_addr = curAddress;
						if(pingStatus.success == true) {
							theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Host was too close! (TTL: %i IP: %s status: %i). Removing this host and restarting host collection.", ttl, inet_ntoa(stDestAddr), pingStatus.status);

							hostsToTraceRoute.RemoveAt(lastPos);
							failed = true;
						} else {
							if(pingStatus.error != IP_REQ_TIMED_OUT) {
								theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Error when pinging a host! (TTL: %i IP: %s Error: %i). Removing this host and restarting host collection. Error info follows.", ttl, inet_ntoa(stDestAddr), pingStatus.error);
								pinger.PIcmpErr(pingStatus.error);

								hostsToTraceRoute.RemoveAt(lastPos);
								failed = true;
							} else {
								theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Timeout when pinging a host! (TTL: %i IP: %s Error: %i). Keeping host. Error info follows.", ttl, inet_ntoa(stDestAddr), pingStatus.error);
								pinger.PIcmpErr(pingStatus.error);
							}
						}
					}
				}

				if(failed == false) {
					if(curHost != 0) {
						if(pingStatus.success && pingStatus.destinationAddress == curHost) {
							IN_ADDR stDestAddr;
							stDestAddr.s_addr = curHost;
							theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Host at TTL %i: %s", ttl, inet_ntoa(stDestAddr));

							lastCommonHost = curHost;
							lastCommonTTL = ttl;
                                    } else if(lastCommonHost != 0 && lastSuccedingPingAddress != 0) {
							foundLastCommonHost = true;
                                        hostToPing = lastSuccedingPingAddress;

                                        IN_ADDR stDestAddr;
                                        stDestAddr.s_addr = hostToPing;
                                        theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Found differing host at TTL %i: %s. This will be the host to ping.", ttl, inet_ntoa(stDestAddr));
						} else {
							foundLastCommonHost = false;
						}
					} else {
						if(ttl < 4) {
							theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Could perform no ping at all at TTL %i. Trying next ttl.", ttl);
						} else {
							theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Could perform no ping at all at TTL %i. Giving up.", ttl);
						}
						lastCommonHost = 0;
					}
				}
		}

                theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Done tracerouting. Evaluating results.");
		if(foundLastCommonHost == true) {
			IN_ADDR stLastCommonHostAddr;
			stLastCommonHostAddr.s_addr = lastCommonHost;

			IN_ADDR stHostToPingAddr;
			stHostToPingAddr.s_addr = hostToPing;

			// log result
			theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Found last common host. LastCommonHost: %s @ TTL: %i HostToPing: %s", inet_ntoa(stLastCommonHostAddr), lastCommonTTL, inet_ntoa(stHostToPingAddr));
		} else {
			theApp.emuledlg->QueueDebugLogLine(false,"SpeedSense: Tracerouting failed to many times. Checking manually entered URL/IP...");
			return false;
		}
	}
	return true;
}