//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#pragma once
#include "types.h"
#include "listensocket.h"
#include "partfile.h"
#include "clientcredits.h"
#include "safefile.h"
#include "BarShader.h"
#include "loggable.h"

enum EUploadState{
	US_UPLOADING,
	US_ONUPLOADQUEUE,
	US_WAITCALLBACK,
	US_CONNECTING,
	US_PENDING,
	US_LOWTOLOWIP,
	US_BANNED,
	US_ERROR,
	US_NONE
};

enum EDownloadState{
	DS_DOWNLOADING,
	DS_ONQUEUE,
	DS_CONNECTED,
	DS_CONNECTING,
	DS_WAITCALLBACK,
	DS_REQHASHSET,
	DS_NONEEDEDPARTS,
	DS_TOOMANYCONNS,
	DS_LOWTOLOWIP,
	DS_BANNED,
	DS_ERROR,
	DS_NONE,
	DS_REMOTEQUEUEFULL  // not used yet, except in statistics
};

enum EChatState{
	MS_NONE,
	MS_CHATTING,
	MS_CONNECTING,
	MS_UNABLETOCONNECT
};

enum EClientSoftware{
	SO_EMULE			= 0,
	SO_CDONKEY			= 1,
	SO_XMULE			= 2,
	SO_SHAREAZA			= 4,
	SO_EDONKEYHYBRID	= 50,
	SO_EDONKEY			= 51,
	SO_MLDONKEY			= 52,
	SO_OLDEMULE			= 53,
	SO_UNKNOWN			= 54
};

enum ESecureIdentState{
	IS_UNAVAILABLE		= 0,
	IS_ALLREQUESTSSEND  = 0,
	IS_SIGNATURENEEDED	= 1,
	IS_KEYANDSIGNEEDED	= 2,
};
enum EInfoPacketState{
	IP_NONE				= 0,
	IP_EDONKEYPROTPACK  = 1,
	IP_EMULEPROTPACK	= 2,
	IP_BOTH				= 3,
};

struct PartFileStamp {
	CPartFile*	file;
	DWORD		timestamp;
};
class CClientReqSocket;
class CFriend;

class CEdt; //<<--enkeyDev(th1) -EDT-

class CUpDownClient: public CLoggable
{
	friend class CUploadQueue;
	friend class CEdt; //<<--enkeyDev(th1) -EDT-
public:
	//base
	CUpDownClient(CClientReqSocket* sender = 0);
	CUpDownClient(uint16 in_port, uint32 in_userid, uint32 in_serverup, uint16 in_serverport,CPartFile* in_reqfile);
	~CUpDownClient();
	void			Disconnected();
	bool			TryToConnect(bool bIgnoreMaxCon = false);
	void			ConnectionEstablished();
	uint32			GetUserID()					{return m_nUserID;}
	void			SetUserID(uint32 nUserID)	{m_nUserID = nUserID;}
	char*			GetUserName()				{return m_pszUsername;}
	uint32			GetIP()						{return m_dwUserIP;}
	bool			HasLowID()					{return (m_nUserID < 16777216);}
	char*			GetFullIP()					{return m_szFullUserIP;}
	uint32			GetUserPort()				{return m_nUserPort;}
	uint32			GetTransferedUp()			{return m_nTransferedUp;}
	uint32			GetTransferedDown()			{return m_nTransferedDown;}
	uint32			GetServerIP()				{return m_dwServerIP;}
	void			SetServerIP(uint32 nIP)		{m_dwServerIP = nIP;}
	uint16			GetServerPort()				{return m_nServerPort;}
	void			SetServerPort(uint16 nPort)	{m_nServerPort = nPort;}
	uchar*			GetUserHash()				{return (uchar*)m_achUserHash;}
	void			SetUserHash(uchar* m_achTempUserHash);
	bool			HasValidHash()				{return ((int*)m_achUserHash[0]) != 0 || ((int*)m_achUserHash[1]) != 0 ||
												        ((int*)m_achUserHash[2]) != 0 || ((int*)m_achUserHash[3]) != 0; }
	int				GetHashType();
	uint32			GetVersion()				{return m_nClientVersion;}
	uint8			GetMuleVersion()			{return m_byEmuleVersion;}
	bool			ExtProtocolAvailable()		{return m_bEmuleProtocol;}
	bool			IsEmuleClient()				{return m_byEmuleVersion;}
	CClientCredits* Credits()					{return credits;}
	bool			IsBanned();
	bool			IsLeecher()					{return m_bLeecher;} //MORPH - Added by IceCream, Anti-leecher feature
	const CString&	GetClientFilename() const	{return m_strClientFilename;}
	bool			SupportsUDP()				{return m_byUDPVer != 0 && m_nUDPPort != 0;}
	uint16			GetUDPPort()				{return m_nUDPPort;}
	void			SetUDPPort(uint16 nPort)	{ m_nUDPPort = nPort; }
	uint8			GetUDPVersion()				{return m_byUDPVer;}
	uint8			GetExtendedRequestsVersion(){return m_byExtendedRequestsVer;}
	uint32			GetIncompletePartVersion()	{return m_incompletepartVer;} //<<-- enkeyDEV(th1) -ICS-
	bool			IsFriend()					{return m_Friend != NULL;}
	float			GetCompression()	{return (float)compressiongain/notcompressed*100.0f;} // Add rod show compression
	void			ResetCompressionGain() {compressiongain = 0; notcompressed=1;} // Add show compression

	void			RequestSharedFileList();
	void			ProcessSharedFileList(char* pachPacket, uint32 nSize, LPCTSTR pszDirectory = NULL);
	// CString			GetUploadFileInfo(); Moved to WebServer.h
	
	void			SetUserName(char* pszNewName);
	EClientSoftware	GetClientSoft()				{return m_clientSoft;}
	void			ReGetClientSoft();
	void			ProcessHelloAnswer(char* pachPacket, uint32 nSize);
	void			ProcessHelloPacket(char* pachPacket, uint32 nSize);
	void			SendHelloAnswer();
	void			SendHelloPacket();
	void			SendMuleInfoPacket(bool bAnswer);
	void			ProcessMuleInfoPacket(char* pachPacket, uint32 nSize);
	void			ProcessMuleCommentPacket(char* pachPacket, uint32 nSize);
	bool			Compare(CUpDownClient* tocomp, bool bIgnoreUserhash = false);
	void			SetLastSrcReqTime()			{m_dwLastSourceRequest = ::GetTickCount();}
	void			SetLastSrcAnswerTime()		{m_dwLastSourceAnswer = ::GetTickCount();}
	void			SetLastAskedForSources()	{m_dwLastAskedForSources = ::GetTickCount();}
	uint32			GetLastSrcReqTime()			{return m_dwLastSourceRequest;}
	uint32			GetLastSrcAnswerTime()		{return m_dwLastSourceAnswer;}
	uint32			GetLastAskedForSources()	{return m_dwLastAskedForSources;}
	DWORD			GetEnteredConnectedState()	{return m_dwEnteredConnectedState;}
	bool			GetFriendSlot();
	void			SetFriendSlot(bool bNV)		{m_bFriendSlot = bNV;}
	void			SetCommentDirty(bool bDirty = true) {m_bCommentDirty = bDirty;}
	uint8			GetSourceExchangeVersion()	{return m_bySourceExchangeVer;}
	// START enkeyDev(th1) -EDT-
	bool			GetDownloadTimeVersion()	{return m_DownloadTimeVer;}
	const CTime &	GetDownloadTime()			{return m_DownloadTime;}
	uint32			GetDownloadTimeVal()		{return m_DownloadTimeVal;}
	uint32			GetDownloadTimeErr()		{return m_DownloadTimeErr;}
	void			EstimateDownloadTime(uint32 &avg_time, uint32 &err_time);
	bool			HasEnkeydevFeatures()		{return m_incompletepartVer || m_DownloadTimeVer;}
	// END enkeyDev(th1) -EDT-

	void			SendPublicKeyPacket();
	void			SendSignaturePacket();
	void			ProcessPublicKeyPacket(uchar* pachPacket, uint32 nSize);
	void			ProcessSignaturePacket(uchar* pachPacket, uint32 nSize);
	uint8			GetSecureIdentState()		{return m_SecureIdentState;}
	void			SendSecIdentStatePacket();
	void			ProcessSecIdentStatePacket(uchar* pachPacket, uint32 nSize);
	void			InfoPacketsReceived();
	void			ResetFileStatusInfo();

	// preview
	void			SendPreviewRequest(CAbstractFile* pForFile);
	void			SendPreviewAnswer(CKnownFile* pForFile, CxImage** imgFrames, uint8 nCount);
	void			ProcessPreviewReq(char* pachPacket, uint32 nSize);
	void			ProcessPreviewAnswer(char* pachPacket, uint32 nSize);
	bool			SupportsPreview()		{return m_bSupportsPreview;}
	bool			SafeSendPacket(Packet* packet);

	CClientReqSocket*	socket;
	CClientCredits*		credits;
	CFriend*			m_Friend;
	//upload
	uint32	compressiongain; /// Add show compression
	uint32  notcompressed; // Add show compression
	EUploadState	GetUploadState()			{return m_nUploadState;}
	void			SetUploadState(EUploadState news);
	uint32			GetWaitStartTime();
	uint32			GetWaitTime()				{return m_dwUploadTime-GetWaitStartTime();}
	bool			IsDownloading()				{return (m_nUploadState == US_UPLOADING);}
	//bool			HasBlocks()					{return !(m_BlockSend_queue.IsEmpty() && m_BlockRequests_queue.IsEmpty());} //Removed by Pawcio
	bool			IsRequest()					{CSingleLock BlockQueueLock(&BlockQueueMutex); BlockQueueLock.Lock(); bool notEmpty=!m_BlockRequests_queue.IsEmpty(); BlockQueueLock.Unlock(); return notEmpty;} //Pawcio: BC
	uint32			GetDatarate()				{return m_nUpDatarate;}	
	uint32			GetScore(bool sysvalue, bool isdownloading = false, bool onlybasevalue = false);
	void			AddReqBlock(Requested_Block_Struct* reqblock);
	bool			m_bAddNextConnect;  // VQB Fix for LowID slots only on connection
	bool			CreateNextBlockPackage();
	void 			SetUpStartTime() {m_dwTimeOnReserved = m_dwUploadTime = ::GetTickCount(); m_dwShieldTime = 0;}//Pawcio: BC
	uint32			GetUpStartTimeDelay()		{return ::GetTickCount() - m_dwUploadTime;}
	void 			SetWaitStartTime();
	void 			ClearWaitStartTime();
	void			SendHashsetPacket(char* forfileid);
	void			SetUploadFileID(uchar* tempreqfileid);
	//Xman see OnUploadqueue
	void			SetOldUploadFileID();
	uchar*			GetOldUploadFileID()		{return oldfileid;} 
	void			SetLastAction(UINT action)	{lastaction=action;}
	UINT			GetLastAction()				{return lastaction;}
	//Xman end
	uchar*			GetUploadFileID()	{return requpfileid;}
	uint32			SendBlockData(uint32 maxammount, bool keepAlive=false); //Pawcio: BC
	void			SplitHeadPacket(); //Maella - //Pawcio: BC
	void			ClearUploadBlockRequests();
	void			SendRankingInfo();
	void			SendCommentInfo(CKnownFile *file);
	uint32			GetLastAskedDelay();
	void			AddRequestCount(uchar* fileid);
	bool 			IsDifferentPartBlock();
	void			UnBan();
	void			Ban();
	void			BanLeecher(int log_message = true); //MORPH - Added by IceCream, Anti-leecher feature
	uint32			GetAskedCount()				{return m_cAsked;}
	void			AddAskedCount()				{m_cAsked++;}
	void			SetAskedCount( uint32 m_cInAsked)				{m_cAsked = m_cInAsked;}
	void			FlushSendBlocks();			// call this when you stop upload, or the socket might be not able to send
	void			SetLastUpRequest()			{m_dwLastUpRequest = ::GetTickCount();}
	uint32			GetLastUpRequest()			{return m_dwLastUpRequest;}
	void			UDPFileReasked();
	uint32			GetSessionUp()			{return m_nTransferedUp - m_nCurSessionUp;}
	void			ResetSessionUp()		{m_nCurSessionUp = m_nTransferedUp; toRemove=false; addOnWaiting=false; toRemoveNow=false; toDelete=false; m_dwLastBlockKeepAliveSent=m_dwLastBlockSent=::GetTickCount();}//Pawcio: BC
	void			ProcessUpFileStatus(char* packet,uint32 size);
	uint16			GetUpPartCount()			{return m_nUpPartCount;}
	void			DrawUpStatusBar(CDC* dc, RECT* rect, bool onlygreyrect, bool  bFlat);

	//MORPH START - Modified by SiRoB, Added by Yun.SF3, ZZ Upload System 20030723-0133
	bool            GetPowerShared();
	//MORPH END - Added by SiRoB, ZZ Upload System 20030723-0133
	//download
	uint32			GetAskedCountDown()				{return m_cDownAsked;} //<<--
	void			AddAskedCountDown()				{m_cDownAsked++;}
	void			SetAskedCountDown( uint32 m_cInDownAsked)				{m_cDownAsked = m_cInDownAsked;}
	EDownloadState	GetDownloadState()			{return m_nDownloadState;}
	void			SetDownloadState(EDownloadState nNewState);
	uint32			GetLastAskedTime()			{return m_dwLastAskedTime;}
	bool			IsPartAvailable(uint16 iPart)	{return	( (iPart >= m_nPartCount) || (!m_abyPartStatus) )? 0:m_abyPartStatus[iPart];} 	
	bool			IsUpPartAvailable(uint16 iPart)	{return	( (iPart >= m_nUpPartCount) || (!m_abyUpPartStatus) )? 0:m_abyUpPartStatus[iPart];} 	
	bool			IsIncPartAvailable(uint16 iPart)	{return	( (iPart >= m_nPartCount) || (!m_abyIncPartStatus) )? 0:m_abyIncPartStatus[iPart];} 	//<<--enkeyDEV(th1) -ICS-
	uint8*			GetPartStatus()				{return m_abyPartStatus;}
	uint16			GetPartCount() const		{return m_nPartCount;}
	uint32			GetDownloadDatarate()		{return m_nDownDatarate;}
	uint16			GetRemoteQueueRank()		{return m_nRemoteQueueRank;}
	void			SetRemoteQueueFull( bool flag )	{m_bRemoteQueueFull = flag;}
	bool			IsRemoteQueueFull()			{return m_bRemoteQueueFull;}
	void			SetRemoteQueueRank(uint16 nr);
	void			SetRemoteEDT(uint32 avg, uint32 err); //<<-- enkeyDEV(th1) -EDT-
	void			DrawStatusBar(CDC* dc, RECT* rect, bool onlygreyrect, bool  bFlat);
	void			AskForDownload();
	void			SendFileRequest();
	void			SendStartupLoadReq();
	void			ProcessFileInfo(char* packet,uint32 size);
	void			ProcessFileStatus(char* packet,uint32 size);
	void			ProcessFileIncStatus(char* packet,uint32 size); //<<-- enkeyDEV(th1) -ICS-
	void			ProcessHashSet(char* packet,uint32 size);
	bool			AddRequestForAnotherFile(CPartFile* file);
	void			SendBlockRequests();
	void			ProcessBlockPacket(char* packet, uint32 size, bool packed = false);
	uint32			CalculateDownloadRate();
	uint16			GetAvailablePartCount();
	bool			SwapToAnotherFile(bool bIgnoreNoNeeded, bool ignoreSuspensions, bool bRemoveCompletely, CPartFile* toFile = NULL);
	void			DontSwapTo(CPartFile* file);
	bool			IsSwapSuspended(CPartFile* file);
	bool			DoSwap(CPartFile* SwapTo, bool anotherfile=false);
	void			UDPReaskACK(uint16 nNewQR);
	void			UDPReaskFNF();
	void			UDPReaskForDownload();
	bool			IsSourceRequestAllowed();
	bool			IsValidSource();
	// -khaos--+++> Download Sessions Stuff
	void			SetDownStartTime()			{m_dwDownStartTime = ::GetTickCount();}
	uint32			GetDownTimeDifference()		{uint32 myTime = m_dwDownStartTime; m_dwDownStartTime = 0; return ::GetTickCount() - myTime;}
	bool			GetTransferredDownMini()	{return m_bTransferredDownMini;}
	void			SetTransferredDownMini()	{m_bTransferredDownMini=true;}
	void			InitTransferredDownMini()	{m_bTransferredDownMini=false;}
	//				A4AF Stats Stuff:
	//				In CPartFile::Process, I am going to keep a tally of how many clients
	//				in that PF's source list are A4AF for other files.  This tally is worthless
	//				to the PartFile that it belongs to, but when we add all of these counts up for
	//				each PartFile, we will get an accurate count of how many A4AF requests there are
	//				total.  This is for the Found Sources section of the tree.  This is a better, faster
	//				option than looping through the lists for unavailable sources.
	uint16			GetA4AFCount()				{return m_OtherRequests_list.GetCount();}
	// <-----khaos-
	
	uint16			GetUpCompleteSourcesCount()	{return m_nUpCompleteSourcesCount;}
	void			SetUpCompleteSourcesCount(uint16 n)	{m_nUpCompleteSourcesCount= n;}

	CPartFile*		reqfile;
	int				sourcesslot;

	//chat
	EChatState		GetChatState()				{return m_nChatstate;}
	void			SetChatState(EChatState nNewS)	{m_nChatstate = nNewS;}
	bool			m_bIsSpammer;
	uint8			m_cMessagesReceived;	// count of chatmessages he sent to me
	uint8			m_cMessagesSend;		// count of chatmessages I sent to him

	//File Comment 
    CString			GetFileComment()			{return m_strComment;} 
    void			SetFileComment(char *desc)	{m_strComment.Format("%s",desc);}
    uint8			GetFileRate()				{return m_iRate;}
    void			SetFileRate(int8 iNewRate)	{m_iRate=iNewRate;}

	// Barry - Process zip file as it arrives, don't need to wait until end of block
	int unzip(Pending_Block_Struct *block, BYTE *zipped, uint32 lenZipped, BYTE **unzipped, uint32 *lenUnzipped, int iRecursion = 0);
	// Barry - Sets string to show parts downloading, eg NNNYNNNNYYNYN
	void ShowDownloadingParts(CString *partsYN);
	void UpdateDisplayedInfo(boolean force=false);
    int             GetFileListRequested() { return m_iFileListRequested; }
    void            SetFileListRequested(int iFileListRequested) { m_iFileListRequested = iFileListRequested; }
	bool			m_bMsgFiltered;

private:
	// base
	void	Init();
	void	ProcessHelloTypePacket(CSafeMemFile* data);
	void	SendHelloTypePacket(CMemFile* data);
	uint32	m_dwUserIP;
	uint32	m_dwServerIP;
	uint32	m_nUserID;
	uint16	m_nUserPort;
	uint16	m_nServerPort;
	uint32	m_nClientVersion;
	uint32	m_nUpDatarate;
	uint32	dataratems;
	uint32	m_cSendblock;
	uint8	m_byEmuleVersion;
	uint8	m_byDataCompVer;
	bool	m_bEmuleProtocol;
	char*	m_pszUsername;
	char	m_szFullUserIP[21];
	char	m_achUserHash[16];
	uint16	m_nUDPPort;
	uint8	m_byUDPVer;
	uint8	m_bySourceExchangeVer;
	uint8	m_byAcceptCommentVer;
	uint8	m_byExtendedRequestsVer;
	uint8	m_cFailed;
	EClientSoftware m_clientSoft;
	uint32	m_dwLastSourceRequest;
	uint32	m_dwLastSourceAnswer;
	uint32	m_dwLastAskedForSources;
    int     m_iFileListRequested;
	bool	m_bFriendSlot;
	bool	m_bCommentDirty;
	bool	m_bIsHybrid;
	bool	m_bIsML;
	bool	m_bGPLEvildoer;
	// preview
	bool	m_bSupportsPreview;
	bool	m_bPreviewReqPending;
	bool	m_bPreviewAnsPending;

	uint32	m_byCompatibleClient;
	CTypedPtrList<CPtrList, Packet*>				 m_WaitingPackets_list;
	CList<PartFileStamp, PartFileStamp>				 m_DontSwap_list;
	DWORD	m_lastRefreshedDLDisplay;
	uint32	m_incompletepartVer;	//<<-- enkeyDEV(th1) -ICS-
	uint32	m_DownloadTimeVer;		//<<-- enkeyDev(th1) -EDT-
	CTime	m_DownloadTime;			//<<-- enkeyDev(th1) -EDT-
	uint32	m_DownloadTimeVal;		//<<-- enkeyDev(th1) -EDT-
	uint32	m_DownloadTimeErr;		//<<-- enkeyDev(th1) -EDT-

	ESecureIdentState	m_SecureIdentState;
	uint8	m_byInfopacketsReceived;			// have we received the edonkeyprot and emuleprot packet already (see InfoPacketsReceived() )
	uint32	m_dwLastSignatureIP;
	uint8	m_bySupportSecIdent;


	//upload
	// -khaos--+++> Added parameters: bool bFromPF = true
	void CreateStandartPackets(byte* data,uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF = true);
	void CreatePackedPackets(byte* data,uint32 togo, Requested_Block_Struct* currentblock, bool bFromPF = true);
	// <-----khaos-
	bool		m_bLeecher; //MORPH - Added by IceCream, anti-leecher feature
	char*		old_m_pszUsername; //MORPH - Added by IceCream, Antileecher feature
	CString		old_m_clientVerString; //MORPH - Added by IceCream, Antileecher feature
	uint32		m_nTransferedUp;
	EUploadState m_nUploadState;
	uint32		m_dwUploadTime;
	uint32		m_nMaxSendAllowed;
	uint32		m_nAvUpDatarate;
	uint32		m_cAsked;
	uint32		m_dwLastUpRequest;
	bool		m_bUsedComprUp;	//only used for interface output
	uint32		m_nCurSessionUp;
	uint16		m_nUpPartCount;
	uint16		m_nUpCompleteSourcesCount;
	static		CBarShader s_UpStatusBar;
	uchar		requpfileid[16];
	uchar		oldfileid[16]; //Xman see OnUploadqueue
	UINT		lastaction; //Xman see OnUploadqueue
	float		cur_ratio;
	float		avg_ratio;
	#define		RATIO_AVG	100
	float		ratios[RATIO_AVG];
	uint8		ratioIndex;
	//Pawcio: BC
	CMutex		BlockQueueMutex;
	//End Pawcio: BC
public:
	uint16		m_lastPartAsked;
	uint8*		m_abyUpPartStatus;
	uint32		m_nSumForAvgUpDataRate;
	CTypedPtrList<CPtrList, CPartFile*>				 m_OtherRequests_list;
	CTypedPtrList<CPtrList, CPartFile*>				 m_OtherNoNeeded_list;
	// Pawcio: BC
	bool		isReady() {return (socket && !socket->IsBusy());}
	bool		hasBlock() {CSingleLock BlockQueueLock(&BlockQueueMutex); BlockQueueLock.Lock(); bool empty=m_BlockSend_queue.IsEmpty(); BlockQueueLock.Unlock(); return !empty;}
	bool		isReadyToSendFileData() {CSingleLock BlockQueueLock(&BlockQueueMutex); BlockQueueLock.Lock(); bool empty=m_BlockSend_queue.IsEmpty(); BlockQueueLock.Unlock(); return (socket && !(socket->IsBusy() || empty || socket->HasControlPacket()));}
	void		ResetProcessRate();
	void		SetLastBlockSent() {m_dwLastBlockSent=::GetTickCount();}
	uint8		processed;
	bool		waitingforparts;
	uint32		m_uBusyTicks;
	uint32		m_uReadyTicks;
	float		GetAvgRatio()		{return avg_ratio;}
	float		GetCurRatio()		{return cur_ratio;}
	void		CalcRatio(bool busy = false);
	bool		toRemove;
	bool		addOnWaiting;
	bool		toRemoveNow;
	bool		toDelete;
	bool		IsToBeRemoved() {CSingleLock BlockQueueLock(&BlockQueueMutex); BlockQueueLock.Lock(); bool empty=m_BlockSend_queue.IsEmpty(); BlockQueueLock.Unlock(); return ((toRemove && empty ||  toRemoveNow || GetTimeSinceLastKeepAliveSent() > CONNECTION_TIMEOUT + 10000)&& !waitingforparts);}
	bool		IsSetToRemove() {return (toRemove || toRemoveNow);}
	uint32		GetPacketSizeToSend();
	void		SetTimeOnReserved() { m_dwTimeOnReserved=::GetTickCount();}
	void		ClearSendHistory() { m_AvarageUDR_list.RemoveAll(); sumavgUDR=0; }
	void		ConsiderReservedTime(bool applyTime=false) { uint32 dwTime = ::GetTickCount(); m_dwUploadTime+=dwTime - m_dwTimeOnReserved; m_dwTimeOnReserved=dwTime; if (applyTime) m_dwLastBlockSent=m_dwLastBlockKeepAliveSent;}
	uint32		GetTimeSinceLastSent()		{ return ::GetTickCount() - m_dwLastBlockSent;}
	uint32		GetTimeSinceLastKeepAliveSent()		{ return ::GetTickCount() - m_dwLastBlockKeepAliveSent;}
	uint32		MinToMB();
	//Pawcio: BC
	const	CString&	GetClientModString() const { return m_clientModString; }//Pawcio: Mod Version (from TBH)
	bool		TestLeecher(); //MORPH - Added by IceCream, anti-leecher feature
// Maella -Extended clean-up II-
	void CleanUp(CPartFile* pDeletedFile);
	DWORD m_lastCleanUpCheck;
// Maella end

private:
	CList<TransferredData,TransferredData>			 m_AvarageUDR_list; // By BadWolf
	CTypedPtrList<CPtrList, Packet*>				 m_BlockSend_queue;
	CTypedPtrList<CPtrList, Requested_Block_Struct*> m_BlockRequests_queue;
	CTypedPtrList<CPtrList, Requested_Block_Struct*> m_DoneBlocks_list;
	CTypedPtrList<CPtrList, Requested_File_Struct*>	 m_RequestedFiles_list;
	//download
	bool		m_bRemoteQueueFull;
	bool		usedcompressiondown; //only used for interface output
	EDownloadState m_nDownloadState;
	uint16		m_nPartCount;
	uint32		m_cDownAsked;
	uint8*		m_abyPartStatus;
	uint8*		m_abyIncPartStatus; //<<-- enkeyDEV(th1) -ICS-
	uint32		m_dwLastAskedTime;
	DWORD		m_dwEnteredConnectedState;
	CString		m_strClientFilename;
	uint32		m_nTransferedDown;
	// -khaos--+++> Download Session Stats
	bool		m_bTransferredDownMini;
	uint32		m_dwDownStartTime;
	// <-----khaos-
	uint32      m_nLastBlockOffset;   // Patch for show parts that you download [Cax2]
	uint32		m_nDownDatarate;
	uint32		m_nDownDataRateMS;
	uint32		m_nAvDownDatarate;
	uint16		m_cShowDR;
	uint32		m_dwLastBlockReceived;
	uint32		m_dwLastBlockSent;		// Pawcio: BC
	uint32		m_dwLastBlockKeepAliveSent;	// Pawcio: BC
	uint32		m_dwTimeOnReserved;		// Pawcio: BC
	uint32		m_dwShieldTime;		// Pawcio: BC
	uint16		m_nRemoteQueueRank;
	bool		m_bCompleteSource;
	bool		m_bReaskPending;
	bool		m_bUDPPending;
	uint32		m_nSumForAvgDownDataRate;

	// By BadWolf - Accurate Speed Measurement (Ottavio84 idea)
	CList<TransferredData,TransferredData>			 m_AvarageDDR_list;
	sint32	sumavgDDR;
	sint32	sumavgUDR;
	// END By BadWolf - Accurate Speed Measurement (Ottavio84 idea)

	CTypedPtrList<CPtrList, Pending_Block_Struct*>	 m_PendingBlocks_list;
	CTypedPtrList<CPtrList, Requested_Block_Struct*> m_DownloadBlocks_list;

	static CBarShader s_StatusBar;
	// chat
	EChatState	m_nChatstate;
	CString		m_strComment;
	int8		m_iRate; 

	// using bitfield for less important flags, to save some bytes
	UINT m_fHashsetRequesting : 1, // we have sent a hashset request to this client
		 m_fSharedDirectories : 1; // client supports OP_ASKSHAREDIRS opcodes
	CString		m_clientModString; // Pawcio: Mod Version (from TBH)
};
